#!/usr/bin/python
from __future__ import print_function

'''
    TSP_TwoOpt.py

    Solves the Traveling Salesman Problem by the Lin Kernighan heuristic.
    Starts with a random path, pairwise edge exchange until no improvements possible.
    As improvement the pair of edges is selected, which gains the best tourlength reduction.
    A description of the generalised K-Opt heuristic can be found in:
      http://de.wikipedia.org/wiki/K-Opt-Heuristik

    Author: Ottmar Kraemer-Fuhrmann <ottmar.kraemer-fuhrmann@scai.fraunhofer.de>

    Adapted to python version 2.x with x>5 and to math module "Optimale Wege"
    by 
    Anton Schueller <anton.schueller@scai.fraunhofer.de>

    $Id$

    Copyright (C) 2012, Fraunhofer SCAI, Germany
    All rights reserved
'''

import os
import random
import sys
import time

import TSP_Utilities

##############
# Initialize #
##############

# Set problem size, i.e. number of cities.
# One second performance: 160 with 1 trial / 40 with 70 trials 
size = input("Number of cities? ")
# size = 40
print("Number of cities:", size)

# Number of trials.
number_of_trials = size
print("Number of trials:", number_of_trials)

# Trace shows actual iteration.
trace = False

###########
# two_opt #
###########
def two_opt(distance):
    '''
        Lin Kernighan heuristic.
        Starts with a random path, improves by pairwise edge exchange.
        As improvement the pair of edges is selected, which gains the best tourlength reduction.

        Input:  distance Quadratic distance matrix.
        Output: Best tour, best tour length.
    '''

    # Random tour.
    tour = [i for i in range(size)]
    random.shuffle(tour)
    if trace:
        print("Initial tour:", TSP_Utilities.tour_length(distance, tour), tour)

    # Pairwise exchange until no improvements possible.
    best_delta = 1
    while (best_delta > 0):
        # Search best exchange pair
        best_delta = 0
        for i in range(size - 3):
            # First path.
            node11 = tour[i]
            node12 = tour[i + 1]
            dist1 = distance[node11][node12]
        
            for j in range(i + 2, size - 1):
                # Second path.
                node21 = tour[j]
                node22 = tour[j + 1]
                dist2 = distance[node21][node22]
            
                # Exchanged distances.
                newDist1 = distance[node11][node21]
                newDist2 = distance[node12][node22]

                # Improvement test.
                delta = dist1 + dist2 - newDist1 - newDist2  
                if (delta > best_delta):
                    # Keep i and j
                    best_delta = delta
                    best_i = i
                    best_j = j
                    
        # Exchange best edge pair
        if best_delta > 0:
            if trace:
                print("Exchange:", -best_delta, " / ", best_i, best_j, " / ",
                      tour[best_i], "-", tour[best_i + 1], " ",
                      tour[best_j], "-", tour[best_j + 1], " in ",
                      TSP_Utilities.tour_length(distance, tour), tour)
            pos1 = best_i + 1
            pos2 = best_j
            while (pos1 < pos2):
                # Exchange cities on tour
                temp = tour[pos1]
                tour[pos1] = tour[pos2]
                tour[pos2] = temp
                # Next segment
                pos1 += 1
                pos2 -= 1
            # print("  -> ", tour_length(distance, tour), tour)
 
    # Compute length of the tour.
    length = TSP_Utilities.tour_length(distance, tour)

    # Return the tour and its length.
    return tour, length

################
# Main program #
################
'''
    Solves the Traveling Salesman Problem by a nearest neighbor search.
             For 2<n<41 a subset of the 40 biggest German towns ist used, for
             n>40 a random distance matrix is generated.
    All nodes are used as starting node and the best tour is selected.
'''

# Read the input matrix, if a distance file exists.
#file_name = "rundreise_input_" + str(size) + ".txt"
file_name = "rundreise_input_40.txt"
if os.path.exists(file_name):
    # Read distance matrix from external file.
    distance = TSP_Utilities.read_distance(file_name,size)
    print("Read distance matrix from file", file_name)
else:
    # Set distance matrix randomly (1 ... 500).
    distance = [[0 for j in range(size)] for i in range(size)]
    for i in range(size):
        distance[i][i] = 0
        for j in range(i + 1, size):
            distance[i][j] = random.randint(1, 500)
            distance[j][i] = distance[i][j]

# Print distance matrix   
if trace:
    TSP_Utilities.print_distance(distance)
print()

# Start timer.
time1 = time.time()

# Try the heuristic several times.
best_length = sys.maxsize
best_tour = []
for trial in range(number_of_trials):
    # Construct new tour.
    tour, length = two_opt(distance)
    if trace:
        print(trial, ") ", length, "", tour)
        
    # Keep best tour found.
    if length < best_length:
        if True:
            print("Better tour (", trial, "): ", length, tour,)
        best_length = length
        best_tour = tour
        
# Shortest tour.
print("\nShortest TwoOpt tour:", best_length, best_tour)

# Stop timer.
time2 = time.time()
print("\nTime:" , time2 - time1)

# Wait for user acknowledge before window disappears.
sys.stdin.readline()
